from TDhelper.generic.transformationType import transformation
from TDhelper.document.excel.FieldType import *
from openpyxl import load_workbook
import csv


class _AttributeOverride:
    def __init__(self, name, m_type):
        self._name = name
        self._type = m_type

    def __get__(self, instance, owen):
        return instance.__dict__[self._name]

    def __set__(self, instance, value):
        instance.__dict__[self._name] = transformation(value, self._type)

    def __delete__(self, instance):
        instance.__dict__.pop(self._name)


class Meta:
    file = None
    sheet = "sheet1"
    extension = "xlsx"
    headrow = True

class modelMeta(type):
    def __new__(cls, name, bases, dct):
        attrs = {
            "mapping": {},
            "__reverse_mapping__": {},
            "__enter__": __enter__,
            "__exit__": __exit__,
            "__rows__": [],
            "close": close,
            "__initExcelHandle__": __initExcelHandle__,
            "getCount": getCount,
            "_translateMode": _translateMode,
            "__rowsCount__": 0,
            "items": items,
            "getitem": getItem,
        }
        for name, value in dct.items():
            if isinstance(dct[name], FieldType):
                attrs["mapping"][name] = value.bindCol
                attrs["__reverse_mapping__"][str(value.bindCol)] = name
                attrs[name] = _AttributeOverride(name, value.fieldType)
            else:
                attrs[name] = value
        return super(modelMeta, cls).__new__(cls, name, bases, attrs)


def __initExcelHandle__(self):
    try:
        if self.Meta.file:
            m_extension = self.Meta.file.rsplit(".")[1]
            if m_extension == "csv":
                self.Meta.extension = "csv"
                self.__excelHandle__ = open(self.Meta.file)
                self.__sheetHandle__ = csv.reader(self.__excelHandle__)
            elif m_extension == "xlsx" or m_extension == "xls":
                self.Meta.extension = "xlsx"
                self.__excelHandle__ = load_workbook(self.Meta.file)
                self.__sheetHandle__ = self.__excelHandle__[self.Meta.sheet]
            else:
                raise Exception("file extension is error.")
            if self.__sheetHandle__:
                for col in self.__sheetHandle__:
                    self.__rows__.append(col)
                if self.Meta.headrow:
                    self.__rows__.pop(0)
                self.__rowsCount__ = len(self.__rows__)
        else:
            raise Exception("meta file is None.")
    except Exception as e:
        raise e


def _translateMode(self, data=[]):
    if data:
        try:
            tmp_index = 1
            for v in data:
                if v == None or v == "None":
                    v = "0"
                if str(tmp_index) in self.__reverse_mapping__:
                    setattr(
                        self,
                        self.__reverse_mapping__[str(tmp_index)],
                        v if not hasattr(v, "value") else v.value,
                    )
                tmp_index += 1
            return self
        except Exception as e:
            raise e
    else:
        raise Exception("translate data is none.")


def items(self):
    assert self.Meta.file, "Meta file is none."
    for col in self.__rows__:
        result = self._translateMode(col)
        yield result


def getItem(self, offset):
    assert 0 <= offset < self.__rowsCount__, "offset %d, " % offset
    return self._translateMode(self.__rows__[offset])

def __enter__(self):
    return self

def __exit__(self, exc_type, exc_value, exc_t):
    self.close()


def close(self):
    self.__excelHandle__.close()
    if self.Meta.extension.lower() != "csv":
        pass#self.__sheetHandle__.close()
    self.__excelHandle__ = None
    self.__sheetHandle__ = None
    self.Meta = None


def getCount(self):
    return self.__rowsCount__



from typing import TypeVar
from TDhelper.db.mongodb.orm.core.attribute import *
from TDhelper.db.mongodb.orm.core.attribute import Any
from TDhelper.db.mongodb.orm.core.field import *
from TDhelper.db.mongodb.orm.drives.conn import mongo_connector,connect_conf,setConf

class objects_cls:
        __delete_state= None
        __drives_cls= None
        __conn_ins= None
        __base_ins= None
        __mongo_queryset= None
        __result_lists= []
        __cursor_pos=0
        @property
        def objects_delete(self):
            return self.__delete_state
        
        @objects_delete.setter
        def objects_delete(self, v:bool):
            if not self.__delete_state:
                self.__delete_state= v
            else:
                raise TypeError("objects_delete is readonly. only can be define once.")
            
        def __init__(self,drive_cls:mongo_connector,ins:Any) -> None:
            self.__drives_cls= drive_cls
            self.__base_ins= ins
            self.__conn_ins= self.__drives_cls(connect_conf).__db__[self.__base_ins.__table__]
            
        def __iter__(self):
            self.__result_lists= list(self.__mongo_queryset)
            return self
        
        def __next__(self) -> object:
            if self.__cursor_pos< len(self.__result_lists):
                o= self.__base_ins(**self.__result_lists[self.__cursor_pos])
                self.__cursor_pos+=1
                return o
            else:
                raise StopIteration()
            
        def all(self):
            return self
        
        def find(self,query={}):
            try:
                self.__mongo_queryset= self.__conn_ins.find(query)
                return self
            except Exception as e:
                raise e
        
        
        def find_for_page(self,query={},pagesize:int=20,page:int=1):
            try:
                page= page if page>0 else 1
                m_skip = (page-1) * pagesize
                if m_skip < 0:
                    m_skip = 0
                self.__mongo_queryset= self.__conn_ins.find(query).limit(pagesize).skip(m_skip)
                return self
            except Exception as e:
                raise e
        
        def skip(self,step:0):
            self.__mongo_queryset= self.__mongo_queryset.skip(step)
            return self
        
        def limit(self,l_c=0):
            self.__mongo_queryset= self.__mongo_queryset.limit(l_c)
            return self
        
        def get(self,pk):
            try:
                return self
            except Exception as e:
                raise e
        
        def insert(self,query) -> Any:
            try:
                return self.__conn_ins.insert_many(query)
            except Exception as e:
                raise e
        
        def update(self,query,set) -> Any:
            try:                
                return self.__conn_ins.update(query,{'$set':set})
            except Exception as e:
                raise e
            
        def delete(self):
            pass
        
T= TypeVar('T')
class new_objects:
    t_cls= None
    args= None
    
    def __init__(self,cls:T, args:tuple=()) -> None:
        self.t_cls= cls
        self.args= args

class mongo_db_meta(type):
    def __new__(cls,name, bases, dct):
        attrs={
            "__fields__":{},
            "__db_drives__":mongo_connector,
            "__table__":name,
            "save": cls.save,
            "update": cls.update,
            "delete": cls.delete
        }
        meta_fields=['table']
        for k,v in dct.items():
            if attrs.__contains__(k):
                continue
            else:
                attrs.update({k:v})
                if isinstance(v,field):
                    if v.model_type != field_type.BsonId:
                        attrs.get("__fields__").update({k.lower():None})
                        attrs[k]=oProperty(k,v)
                    else:
                        if v.default:
                            attrs.get("__fields__").update({k.lower():bson.objectid.ObjectId(v.default)})
                        else:
                            attrs.get("__fields__").update({k.lower():None})
                        attrs[k]=oProperty(k,v)
                if k == "Meta":
                    for o in meta_fields:
                        if not hasattr(v,o):
                            raise ValueError("Can not set Meta attribute '%s'" % o)
                        else:
                            attrs["".join(["__",o,"__"])]= v.__dict__[o]
        if not attrs['__fields__'].__contains__("_id"):
            attrs["_id"]=oProperty("_id",field_type.BsonId)
        new_cls= super(mongo_db_meta,cls).__new__(cls,name,bases,attrs) 
        setattr(cls,"_objects_ins",new_objects(objects_cls,(attrs["__db_drives__"],new_cls)))
        return new_cls

    def save(cls):
        try:
            return cls.__collect__.insert_one(cls.__fields__)
        except Exception as e:
            raise e
            
    def update(cls):
        try:
            sets= {}
            # exclude bson field.
            for k,v in cls.__fields__.items():
                if not isinstance(v,bson.objectid.ObjectId):
                    sets[k]=v
            result= cls.__collect__.update_one({"_id":cls._id},{"$set":sets})
            return result
        except Exception as e:
            cls.__fields__
            raise e
    
    def delete(cls):
        try:
            _id= cls._id
            cls.__fields__.clear()
            return cls.__collect__.delete_one({"_id":_id})
        except Exception as e:
            raise e
        
    
    @classmethod
    @property
    def objects(cls) -> objects_cls:
        return cls._objects_ins.t_cls(*cls._objects_ins.args)
    


from TDhelper.network.http.REST_HTTP import POST, GET
from datetime import datetime, timedelta, timezone
import json


class token_manage:
    __token = None
    __expire = None
    __token_service_uri = None
    __service_key = None
    __service_secret = None
    __time_cls = None

    @property
    def Token(self):
        try:
            if not self.__token:
                self.__get_token__()
            if self.__expire:
                if datetime.strptime(self.__expire,"%Y-%m-%d %H:%M:%S.%f") < self.__time_cls.now():
                    self.__get_token__()
            return self.__token
        except Exception as e:
            return ""

    def __init__(self, token_uri:str, svr_key:str, secret:str, time_cls=None, path:str= None, post_data:dict={}) -> None:
        self.__time_cls = datetime if not time_cls else time_cls
        self.__service_key = svr_key
        self.__service_secret = secret
        token_uri= token_uri.rstrip('/')
        self.__token_service_uri = "/".join([token_uri, path if path else "api/rpc_token"])
        self.__post_data= post_data if post_data else {
                "svr_key": self.__service_key.upper(),
                "secret": self.__service_secret
                }
        #self.__get_token__()

    def __get_token__(self):
        try:
            state, body = GET(
                self.__token_service_uri,
                self.__post_data,
                time_out=5
            )
            if state:
                data = json.loads(body)
                if data.get("state")==200:
                    self.__token = data.get("msg").get("token", None)
                    self.__expire = data.get("msg").get("expire", None).replace("T"," ")
                else:
                    raise Exception(data.get('msg'))
        except Exception as e:
            raise e



import json
#! python 3.6
import json
from typing import Any

class Result:

    __result_code__ = 0
    __result_status__ = False
    __result_message = ""
    __result_data = Any

    def __init__(self, code: int, status: bool, message: str, data: dict | str) -> object:
        """__init__
            Instance Result object.
        Parameters:
            code - <class:int>, result code.
            status - <class:bool>, result status.
            message - <class:str>, if status is false, then read message get the error message.
            data - <class:str|dict>, if status is true, the read data get the data.
        Returns:
            <class: object>, the result instance.
        """
        self.__result_code__ = code
        self.__result_status__ = status
        self.__result_message__ = message
        self.__result_data__ = data

    @classmethod
    def dumps(self,isJson= False) -> json:
        """jump
            jump instance to json object.
        Parameters:
            self - Result instance.
            isJson - control result type. default False, return Dict, can set True then will return Json str.
        Returns:
            <class: dict|json>, result dict or json.
        """
        m_result= {
            "code": self.__result_code__,
            "status": self.__result_code__,
            "message": self.__result_message,
            "data": self.__result_data
        }
        if isJson:
            m_result=json.dumps(m_result)
        return m_result

    @classmethod
    def loads(self,_json:str|dict) -> object:
        if isinstance(_json,str):
            _json= json.loads(_json)
        if _json:
            if 'code' not in _json:
                raise(Exception("Missing field 'code'."))
            if 'status' not in _json:
                raise(Exception("Missing field 'status'."))
            if _json['status']:
                if 'data' not in _json:
                    raise(Exception("Missing field 'data'"))
            else:
                if 'message' not in _json:
                    raise(Exception("Missing field 'message'."))
            return Result(_json['code'],_json['status'],_json['message'],_json['data'])



from types import FunctionType, MethodType, ModuleType

class eventMeta(type):
    def __new__(cls, name, bases, dct):
        attrs = {'_eventList': {}}
        for key, value in dct.items():
            attrs[key]= value
        attrs['_event_register']= register
        attrs['_event_on']= trigger
        attrs['_event_remove']= removeEvent
        return super(eventMeta, cls).__new__(cls, name, bases, attrs)

def register(self, name, func):
    if name not in self._eventList:
        self._eventList[name] = func
    else:
        raise Exception('event (%s) already registered.' % name)


def removeEvent(self, name):
    if name in self._eventList:
        self._eventList.pop(name)
    else:
        raise Exception('event (%s) not registered.' % name)


def trigger(self, name, *args, **kwargs):
    if name in self._eventList:
        self._eventList[name](self, *args, **kwargs)
    else:
        raise Exception('event (%s) not registered.' % name)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pymongo
from urllib.parse import quote_plus
'''
class\r\n
    lib.db.mongodb.mongodbclient\r\n
description\r\n
    Mongodb database helper class\r\n
'''
connect_conf={
    "host":"",
    "port":0,
    "db":"",
    "usr":"",
    "pw":""
}

def setConf(host:str,port:int,db:str,usr:str="",pw:str=""):
    connect_conf["host"]= host
    connect_conf["port"]= port
    connect_conf["db"]= db
    connect_conf["usr"]= usr if usr else ""
    connect_conf["pw"]= pw if pw else ""

class mongo_connector:
    __client__= None
    __db__= None
    __db_name__=""
    __url__= ""
    def __init__(self, conf:dict):
        if conf["usr"] and conf["pw"]:
            self.__url__= "mongodb://%s:%s@%s:%d" % (conf["usr"],conf["pw"],conf["host"],conf["port"])
        else:
            self.__url__= "mongodb://%s:%d" % (conf["host"],conf["port"])
        if self.__url__:
            self.__db_name__= conf["db"]
            self.__client__=pymongo.MongoClient(self.__url__,connect=False)
            self.__db__= self.__client__[conf["db"]]
        else:
            raise Exception("mongodb url is none.")
        
    def __get_new_client__(self):
        if self.__url__:
            return pymongo.MongoClient(self.__url__)[self.__db_name__]
        else:
            raise Exception("mongodb url is none.")




class Scheduler():
    def __init__(self,name,plugin,taskType="interval",startTime=None,sleep=0):
        """Init task
            params:
                name: str
                plugin:call task module
                taskType:task type
                    realtime:run at now
                    interval:loop run and sleep N millisecond,default value
                    timing:run at one datetime
                startTime:first run time
                sleep:sleep millisecond
        """
        self.__config={
        'name':name,
        'plugin':plugin,
        'type':taskType,#realtime,interval,timing
        'starttime':startTime,
        'sleep':sleep,#millisecond;
    }

    def Config(self):
        return self.__config
